# Copyright 2025 Ant Group Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Analysis and visualization utilities for mplang.

This subpackage hosts non-core developer aids: diagram rendering, IR dumps,
profiling helpers (future), etc.
"""

from mplang.analysis.diagram import (
    DumpResult,
    FlowchartOptions,
    SequenceDiagramOptions,
    dump,
    to_flowchart,
    to_sequence_diagram,
)

__all__ = [
    "DumpResult",
    "FlowchartOptions",
    "SequenceDiagramOptions",
    "dump",
    "to_flowchart",
    "to_sequence_diagram",
]
